

import java.util.*;



public class Parser {
	public static final int _EOF = 0;
	public static final int _number = 1;
	public static final int _name = 2;
	public static final int _nameabbr = 3;
	public static final int maxT = 12;

	static final boolean T = true;
	static final boolean x = false;
	static final int minErrDist = 2;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;
	
	public Scanner scanner;
	public Errors errors;

	

	public Parser(Scanner scanner) {
		this.scanner = scanner;
		errors = new Errors();
	}

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (String msg) {
		if (errDist >= minErrDist) errors.SemErr(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) {
				++errDist;
				break;
			}

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	boolean StartOf (int s) {
		return set[s][la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}
	
	boolean WeakSeparator (int n, int syFol, int repFol) {
		int kind = la.kind;
		if (kind == n) { Get(); return true; }
		else if (StartOf(repFol)) return false;
		else {
			SynErr(n);
			while (!(set[syFol][kind] || set[repFol][kind] || set[0][kind])) {
				Get();
				kind = la.kind;
			}
			return StartOf(syFol);
		}
	}
	
	void Phonebook() {
		Phonebook phonebook = new Phonebook(); 
		while (la.kind == 2) {
			Entry e; 
			e = Entry();
			phonebook.entries.add(e); 
		}
	}

	Entry  Entry() {
		Entry  result;
		String lastname;
		List<String> firstnames = new ArrayList<>();
		List<PhoneNumber> numbers = new ArrayList<>(); 
		Expect(2);
		lastname = t.val; 
		Expect(4);
		while (la.kind == 2 || la.kind == 3) {
			if (la.kind == 2) {
				Get();
			} else {
				Get();
			}
			firstnames.add(t.val); 
		}
		while (StartOf(1)) {
			PhoneNumber number; 
			number = PhoneNumber();
			numbers.add(number); 
		}
		result = new Entry(lastname, firstnames, numbers); 
		return result;
	}

	PhoneNumber  PhoneNumber() {
		PhoneNumber  result;
		PhoneNumber.Type type = null; 
		if (la.kind == 5 || la.kind == 6 || la.kind == 7) {
			if (la.kind == 5) {
				Get();
				type = PhoneNumber.Type.HOME; 
			} else if (la.kind == 6) {
				Get();
				type = PhoneNumber.Type.OFFICE; 
			} else {
				Get();
				type = PhoneNumber.Type.MOBILE; 
			}
		}
		long country = 0; long area = 0; 
		if (la.kind == 9 || la.kind == 10 || la.kind == 11) {
			if (la.kind == 10 || la.kind == 11) {
				country = Country();
				Expect(8);
				Expect(1);
				area = Long.parseLong(t.val); 
			} else {
				Get();
				Expect(1);
				area = Long.parseLong(t.val); 
			}
		}
		List<Long> numbers = new ArrayList<>(); 
		Expect(1);
		numbers.add(Long.parseLong(t.val)); 
		while (la.kind == 1) {
			Get();
			numbers.add(Long.parseLong(t.val)); 
		}
		result = new PhoneNumber(type, country, area, numbers); 
		return result;
	}

	long  Country() {
		long  result;
		if (la.kind == 10) {
			Get();
		} else if (la.kind == 11) {
			Get();
		} else SynErr(13);
		Expect(1);
		result = Long.parseLong(t.val); 
		return result;
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		Phonebook();
		Expect(0);

	}

	private static final boolean[][] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x},
		{x,T,x,x, x,T,T,T, x,T,T,T, x,x}

	};


	public String ParseErrors(){
			java.io.PrintStream oldStream = System.out;
		
		java.io.OutputStream out = new java.io.ByteArrayOutputStream();
		java.io.PrintStream newStream = new java.io.PrintStream(out);
		
		errors.errorStream = newStream;
		
		Parse();

		String errorStream = out.toString();
		errors.errorStream = oldStream;

		return errorStream;


	}
} // end Parser


class Errors {
	public int count = 0;                                    // number of errors detected
	public java.io.PrintStream errorStream = System.out;     // error messages go to this stream
	public String errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
	
	protected void printMsg(int line, int column, String msg) {
		StringBuffer b = new StringBuffer(errMsgFormat);
		int pos = b.indexOf("{0}");
		if (pos >= 0) { b.delete(pos, pos+3); b.insert(pos, line); }
		pos = b.indexOf("{1}");
		if (pos >= 0) { b.delete(pos, pos+3); b.insert(pos, column); }
		pos = b.indexOf("{2}");
		if (pos >= 0) b.replace(pos, pos+3, msg);
		errorStream.println(b.toString());
	}
	
	public void SynErr (int line, int col, int n) {
		String s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "number expected"; break;
			case 2: s = "name expected"; break;
			case 3: s = "nameabbr expected"; break;
			case 4: s = "\",\" expected"; break;
			case 5: s = "\"home\" expected"; break;
			case 6: s = "\"office\" expected"; break;
			case 7: s = "\"mobile\" expected"; break;
			case 8: s = "\"(0)\" expected"; break;
			case 9: s = "\"0\" expected"; break;
			case 10: s = "\"+\" expected"; break;
			case 11: s = "\"00\" expected"; break;
			case 12: s = "??? expected"; break;
			case 13: s = "invalid Country"; break;
			default: s = "error " + n; break;
		}
		printMsg(line, col, s);
		count++;
	}

	public void SemErr (int line, int col, String s) {	
		printMsg(line, col, s);
		count++;
	}
	
	public void SemErr (String s) {
		errorStream.println(s);
		count++;
	}
	
	public void Warning (int line, int col, String s) {	
		printMsg(line, col, s);
	}
	
	public void Warning (String s) {
		errorStream.println(s);
	}
} // Errors


class FatalError extends RuntimeException {
	public static final long serialVersionUID = 1L;
	public FatalError(String s) { super(s); }
}
